package k.ld31;

import flash.Lib;
import flash.display.Bitmap;
import flash.display.BitmapData;
import flash.display.Shape;
import flash.display.Sprite;
import flash.events.Event;
import flash.events.KeyboardEvent;
import flash.events.MouseEvent;
import flash.text.TextField;
import flash.text.TextFormat;
import flash.text.TextFormatAlign;
import flash.text.TextFieldAutoSize;
import flash.ui.Keyboard;

import k.lib.graphics.Assets;
import k.lib.graphics.DisplayObject;
import k.lib.graphics.RenderEngine;
import k.lib.graphics.Shake;

import mt.deepnight.PathFinder;

typedef Point = {
	var x:Int;
	var y:Int;
}

typedef LevelData = {
	var walls:Array<Array<Bool>>;
}

@:enum abstract CellColor(Int) {
	var empty = 0xFFFFFF;
	var wall = 0x0;
	var tama = 0xFF00FF;
	var cookie = 0x00FF00;
	var player = 0xFFFF00;
	var spawn = 0xFF0000;
	var spawn_unique = 0x0000FF;
}

enum InputNeeded {
	FOOD;
	HELP;
	INSTRUCTIONS_READ;
}

enum GameState {
	LOADING;
	LOADED;
	CINEMATIC(i:Int);
	WAIT_INPUT(i:InputNeeded);
	ALMOST_PLAYING;
	PLAYING;
	GAME_OVER;
}

class Tama extends Sprite {
	public static var instance:Tama;
	public static var frame:Int;
	public static var state(default, null):GameState;
	public static var locked:Bool;

	public static inline var CELL_SIZE:Int = 15;
	public static inline var PLATEAU_X:Int = 150;
	public static inline var PLATEAU_Y:Int = 210;
	public static inline var CINEMATIC_WAIT:Int = 90;
	private static var MAX_LEVEL(default, null):Int = 0;
	private static var WALLS_PER_ITERATION(default, null):Int = 1;

	private var render:RenderEngine;
	public var gameGraphics(default, null):GraphicalInterface;
	private var bitmap:Bitmap;
	private var bd:BitmapData;
	private var _width:Int;
	private var _height:Int;
	private var forceUpdate:Bool;
	private var currLevel:Int;
	public var cinematic_start:Int;

	// Entities
	public var tama:TamaEntity;
	public var player:Player;
	public var walls:Array<DisplayObject>;
	public var poo:Array<DisplayObject>;
	public var wallInfo:Array<Array<Bool>>;
	public var target:DisplayObject;
	public var cookie:DisplayObject;
	public var shurikens:DisplayObject;
	public var tempDO:DisplayObject;
	public var barLeft:Bar;
	public var barRight:Bar;
	public var monsters:Array<Monster>;

	public var newWalls:Array<Point>;
	public var oldWalls:Array<Point>;

	public var pathfinder:PathFinder;

	public function new() {
		super();
		state = LOADING;
		instance = this;

		addEventListener(Event.ADDED_TO_STAGE, added);
	}

	public function added(?d:Dynamic):Void {
		removeEventListener(Event.ADDED_TO_STAGE, added);
		state = LOADED;
		locked = true;

		// Hard-coded dimensions
		_width = 675; _height = 780;

		render = new RenderEngine(_width, _height);
		render.setFlag("alpha", true);
		render.setFlag("rgb", true);
		render.setFlag("rotation", true);
		// render.setFlag("scale", true);

		// Prepare assets
		render.add(Assets.getBitmapData("assets/boom-A1.png"), "boom-A1");
		render.add(Assets.getBitmapData("assets/boom-A2.png"), "boom-A2");
		render.add(Assets.getBitmapData("assets/boom-B1.png"), "boom-B1");
		render.add(Assets.getBitmapData("assets/boom-B2.png"), "boom-B2");
		render.add(Assets.getBitmapData("assets/boom-C1.png"), "boom-C1");
		render.add(Assets.getBitmapData("assets/boom-C2.png"), "boom-C2");
		render.add(Assets.getBitmapData("assets/boom-D1.png"), "boom-D1");
		render.add(Assets.getBitmapData("assets/boom-D2.png"), "boom-D2");

		render.add(Assets.getBitmapData("assets/givecandy.png"), "givecandy");
		render.add(Assets.getBitmapData("assets/gohelp.png"), "gohelp");
		render.add(Assets.getBitmapData("assets/howto.png"), "howto");
		render.add(Assets.getBitmapData("assets/thanks.png"), "thanks");

		render.add(Assets.getBitmapData("assets/tama.png"), "tama");
		render.add(Assets.getBitmapData("assets/wall.png"), "wall");
		render.add(Assets.getBitmapData("assets/player.png"), "player");
		render.add(Assets.getBitmapData("assets/cookie.png"), "cookie");
		render.add(Assets.getBitmapData("assets/monster.png"), "monster");
		render.add(Assets.getBitmapData("assets/shuriken.png"), "shuriken");
		render.add(Assets.getBitmapData("assets/poo.png"), "poo");
		render.add(Assets.getBitmapData("assets/bar.png"), "bar");
		render.add(Assets.getBitmapData("assets/grid.png"), "grid");
		render.add(Assets.getBitmapData("assets/gameover.png"), "gameover");
		render.add(Assets.getBitmapData("assets/mask-left.png"), "mask-left");
		render.add(Assets.getBitmapData("assets/mask-right.png"), "mask-right");
		render.add(Assets.getBitmapData("assets/background.png"), "background");
		TamaEntity.prepareTexts(render);
		render.pack();

		#if (flash && !FORCE_TILES)
		bd = new BitmapData(_width, _height, false, 0x00FFFFFF);
		bitmap = new Bitmap(bd);
		addChild(bitmap);
		#end
		
		#if DEBUG Lib.current.stage.addChild(new openfl.display.FPS(10,10,0xFF0000)); #end
		Lib.current.addEventListener(Event.ENTER_FRAME, update);
		Lib.current.addEventListener(MouseEvent.MOUSE_DOWN, lock);
		Lib.current.addEventListener(MouseEvent.MOUSE_UP, unlock);
		Lib.current.stage.addEventListener(KeyboardEvent.KEY_DOWN, onkey);

		// Ensure that the stage has focus
		Lib.current.stage.focus = Lib.current.stage;

		launch();
	}

	@:access(k.lib.graphics.DisplayObject)
	public function onkey(e:KeyboardEvent):Void {
		switch (state) {
			case GAME_OVER:
				if (e.keyCode == Keyboard.ENTER) launch();

			case WAIT_INPUT(i):
				if (e.keyCode == Keyboard.ENTER) {
					switch (i) {
						case FOOD:
						gameGraphics.top.removeChild(tempDO);

						var x:Int = 8;
						var y:Int = 10;
						cookie = new DisplayObject("cookie");
						cookie.__align = DisplayObject.ALIGN_CENTER;
						cookie.cx = x;
						cookie.cy = y;
						cookie.x = toAbsoluteX(x);
						cookie.y = toAbsoluteY(y);
						gameGraphics.mid.addChild(cookie);
						state = ALMOST_PLAYING;
						locked = false;

						case HELP:
						gameGraphics.top.removeChild(tempDO);
						var x:Int = 10;
						var y:Int = 16;
						player = new Player(toAbsoluteX(x), toAbsoluteY(y)); gameGraphics.midtop.addChild(player);
						state = ALMOST_PLAYING;
						locked = false;

						case INSTRUCTIONS_READ:
						// gameGraphics.midtop.removeChild(player);
						// gameGraphics.midtop.removeChild(tama);
						// player = null; tama = null;
						gameGraphics.top.removeChild(tempDO);
						var newLevel:LevelData = loadLevel(levelString(currLevel));
						updateWalls(newLevel);
						state = PLAYING;
					}
				}

			case _:
			// Nothing
		}
	}

	private function launch():Void {
		Spawner.clear();
		newWalls = [];
		oldWalls = [];
		monsters = [];
		poo = [];
		frame = 0;
		cinematic_start = 0;
		tama = null;
		player = null;
		cookie = null;

		if (gameGraphics == null) gameGraphics = new GraphicalInterface(_width, _height);
		else gameGraphics.clearAll();
		gameGraphics.back.addChild(new DisplayObject("background"));

		barLeft = new Bar(75, 315, 178, 0.0);
		gameGraphics.top.addChild(barLeft);

		barRight = new Bar(575, 315, 178, 1.0);
		gameGraphics.top.addChild(barRight);
		
		var maskLeft:DisplayObject = new DisplayObject("mask-left");
		maskLeft.x = 60;
		maskLeft.y = 315;
		gameGraphics.top.addChild(maskLeft);
		var maskRight:DisplayObject = new DisplayObject("mask-right");
		maskRight.x = 570;
		maskRight.y = 315;
		gameGraphics.top.addChild(maskRight);

		// Init player target
		target = new DisplayObject(); //"target");
		target.x = -1;
		target.y = -1;
		gameGraphics.top.addChild(target);

		MAX_LEVEL = 0;
		while (Assets.getBitmapData(levelString(MAX_LEVEL+1)) != null) MAX_LEVEL++;

		pathfinder = new PathFinder(25, 25);
		pathfinder.cacheSymetricPaths = false;
		currLevel = 1;
		var level:LevelData = loadLevel(levelString(currLevel));
		walls = new Array<DisplayObject>();
		updateWalls(level, true);

		shurikens = new DisplayObject();
		gameGraphics.mid.addChild(shurikens);

		forceUpdate = true;
		
		// Ensure that the stage has focus
		Lib.current.stage.focus = Lib.current.stage;

		tama.talk(0);
		state = CINEMATIC(0);
	}

	private function lock(d:Dynamic):Void {
		if (target.x != -1) {
			player.lock = true;
		}
	}

	private function unlock(d:Dynamic):Void {
		player.lock = false;
	}

	private inline function update(?d:Dynamic):Void {
		var changed:Bool = preUpdate();

		if (changed || forceUpdate) {
			#if (flash && !FORCE_TILES)
			if (bd != null) render.renderBitmapData(bd, gameGraphics);
			#else
			render.renderTiles(graphics, gameGraphics, true, false);
			#end
		}

		forceUpdate = false;
	}

	@:access(k.lib.graphics.DisplayObject)
	private function preUpdate():Bool {
		frame++;

		switch (state) {
			case LOADING: return false;
			case LOADED: return true;

			case CINEMATIC(x):
			switch (x) {
				case 0:
				if (frame - cinematic_start > CINEMATIC_WAIT) {
					tama.talk(6);
					state = WAIT_INPUT(FOOD);
					tempDO = new DisplayObject("givecandy");
					tempDO.x = toAbsoluteX(12);
					tempDO.y = toAbsoluteY(20);
					tempDO.__align = DisplayObject.ALIGN_CENTER;
					gameGraphics.top.addChild(tempDO);
				}

				case 1:
				if (frame - cinematic_start > CINEMATIC_WAIT) {
					tama.talk(6);
					state = CINEMATIC(2);
					cinematic_start = frame;
				}

				case 2:
				// if (frame - cinematic_start > 120) {
				// 	tama.talk(7);
				// 	state = CINEMATIC(3);
				// 	cinematic_start = frame;
				// }

				// case 3:
				if (frame - cinematic_start > CINEMATIC_WAIT) {
					tama.talk(8);
					state = WAIT_INPUT(HELP);
					tempDO = new DisplayObject("gohelp");
					tempDO.x = toAbsoluteX(12);
					tempDO.y = toAbsoluteY(20);
					tempDO.__align = DisplayObject.ALIGN_CENTER;
					gameGraphics.top.addChild(tempDO);
				}
			}

			case ALMOST_PLAYING, PLAYING:
			// Update player target
			if (player != null && target != null) {
				if (target.x == -1 || target.y == -1) {
					var xDistance:Float = Lib.current.stage.mouseX - player.customx;
					var yDistance:Float = Lib.current.stage.mouseY - player.customy;

					if (Math.sqrt(yDistance*yDistance + xDistance*xDistance) < CELL_SIZE * 8) {
						target.x = Std.int(Lib.current.stage.mouseX);
						target.y = Std.int(Lib.current.stage.mouseY);
					}
				} else {
					var speed:Float = 0.12;
					// var xDistance:Float = Lib.current.stage.mouseX - target.customx;
					// var yDistance:Float = Lib.current.stage.mouseY - target.customy;
					// target.customx -= Std.int((target.customx - Lib.current.stage.mouseX) * speed);
					// target.customy -= Std.int((target.customy - Lib.current.stage.mouseY) * speed);
					var xDistance:Float = Lib.current.stage.mouseX - target.x;
					var yDistance:Float = Lib.current.stage.mouseY - target.y;
					target.x -= Std.int((target.x - Lib.current.stage.mouseX) * speed);
					target.y -= Std.int((target.y - Lib.current.stage.mouseY) * speed);

					// Snap to grid
					// var cell_x:Int = toCellX(target.customx);
					// var cell_y:Int = toCellY(target.customy);
					// if (cell_x < 1) cell_x = 1;
					// if (cell_x > 23) cell_x = 23;
					// if (cell_y < 1) cell_y = 1;
					// if (cell_y > 23) cell_y = 23;
					// if (!wallInfo[cell_x][cell_y]) {
					// 	target.x = cell_x * CELL_SIZE + PLATEAU_X;
					// 	target.y = cell_y * CELL_SIZE + PLATEAU_Y;
					// }
					if (target.x < CELL_SIZE + PLATEAU_X) target.x = CELL_SIZE + PLATEAU_X;
					if (target.y < CELL_SIZE + PLATEAU_Y) target.y = CELL_SIZE + PLATEAU_Y;
					if (target.x > CELL_SIZE * 23 + PLATEAU_X) target.x = CELL_SIZE * 23 + PLATEAU_X;
					if (target.y > CELL_SIZE * 23 + PLATEAU_Y) target.y = CELL_SIZE * 23 + PLATEAU_Y;
				}
			}

			// Display new level
			if (newWalls.length > 0 && frame % 2 == 0) {
				for (i in 0...WALLS_PER_ITERATION) {
					var wallPoint:Point = newWalls.shift();
					var wall:DisplayObject = new DisplayObject("wall");
					wall.x = wallPoint.x * CELL_SIZE + PLATEAU_X;
					wall.y = wallPoint.y * CELL_SIZE + PLATEAU_Y;
					wall.cx = wallPoint.x;
					wall.cy = wallPoint.y;
					gameGraphics.mid.addChild(wall);
				}
			}
			if (oldWalls.length > 0 && frame % 2 == 1) {
				for (i in 0...WALLS_PER_ITERATION) {
					var wall:Point = oldWalls.shift();
					
					var children:Array<DisplayObject> = gameGraphics.mid.getChildren();
					for (c in children) {
						if (c.cx == wall.x && c.cy == wall.y) {
							gameGraphics.mid.removeChild(c);
							break;
						}
					}
				}
			}

			if (newWalls.length == 0 && oldWalls.length == 0) {
				locked = false;
			}

			case _:
			// Nothing to do?

		}

		gameGraphics.update();
		return true; // Always update
	}

	public static inline function toCellX(x:Float):Int return Math.floor((x - PLATEAU_X) / CELL_SIZE);
	public static inline function toCellY(y:Float):Int return Math.floor((y - PLATEAU_Y) / CELL_SIZE);
	public static inline function toAbsoluteX(x:Int):Int return x * CELL_SIZE + PLATEAU_X + Std.int(CELL_SIZE/2);
	public static inline function toAbsoluteY(y:Int):Int return y * CELL_SIZE + PLATEAU_Y + Std.int(CELL_SIZE/2);

	public static inline function canPass(x:Int, y:Int):Bool {
		if (x < 0 || y < 0 || x > 24 || y > 24) return false;
		return !instance.wallInfo[x][y];
	}

	public static inline function hitPlayer(monster:Monster):Void {
		var tama:TamaEntity = instance.tama;
		var player:Player = instance.player;
		tama.life--;
		instance.barRight.updateValue(tama.life / TamaEntity.MAX_LIFE);
		player.shake(Shake.SHAKE_NEUTRAL);

		if (tama.life <= 0) {
			var b:Boom = new Boom();
			b.x = player.x;
			b.y = player.y;
			instance.gameGraphics.midtop.addChild(b);
			instance.gameOver();
		} else {
			tama.talk(TamaEntity.text_playerhit[Std.int(Math.random() * TamaEntity.text_playerhit.length)]);

			var b:Boom = new Boom();
			b.x = monster.x;
			b.y = monster.y;
			instance.gameGraphics.midtop.addChild(b);

			killMonster(monster);
		}
	}

	public static inline function hitTama(monster:Monster):Void {
		var tama:TamaEntity = instance.tama;
		tama.life--;
		instance.barRight.updateValue(tama.life / TamaEntity.MAX_LIFE);
		tama.shake(Shake.SHAKE_NEUTRAL);

		if (tama.life <= 0) {
			var b:Boom = new Boom();
			b.x = tama.x;
			b.y = tama.y;
			instance.gameGraphics.midtop.addChild(b);
			instance.gameOver();
		} else {
			tama.talk(TamaEntity.text_hit[Std.int(Math.random() * TamaEntity.text_hit.length)]);

			var b:Boom = new Boom();
			b.x = monster.x;
			b.y = monster.y;
			instance.gameGraphics.midtop.addChild(b);

			killMonster(monster);
		}
	}

	public static inline function killMonster(monster:Monster):Void {
		var b:Boom = new Boom();
		b.x = monster.x;
		b.y = monster.y;
		instance.gameGraphics.midtop.addChild(b);

		instance.monsters.remove(monster);
		instance.gameGraphics.midtop.removeChild(monster);
	}

	public static inline function hitMonster(monster:Monster):Void {
		monster.life--;
		monster.shake(Shake.SHAKE_NEUTRAL);
		if (monster.life <= 0) killMonster(monster);
	}

	public static inline function killShuriken(shuriken:DisplayObject):Void {
		// TODO: animation
		instance.shurikens.removeChild(shuriken);
	}

	public function gameOver():Void {
		// TODO: explode player, tama & monsters

		locked = true;
		state = GAME_OVER;
		gameGraphics.fade(gameOverScreen);
	}

	public function gameOverScreen():Void {
		gameGraphics.mid.clear();
		gameGraphics.midtop.clear();
		gameGraphics.top.clear();

		// TODO
		var d:DisplayObject = new GameOver();
		gameGraphics.top.addChild(d);
	}


	private inline function levelString(level:Int):String {
		if (level > 9) return 'assets/level-$level.png';
		return 'assets/level-0$level.png';
	}

	@:access(k.lib.graphics.DisplayObject)
	public function nextLevel():Void {
		locked = true;
		barLeft.updateValue(currLevel / MAX_LEVEL);
		currLevel++;

		if (currLevel > MAX_LEVEL) {
			Spawner.clear();
			gameGraphics.mid.removeChild(cookie);
			for (m in monsters.copy()) if (m != null) killMonster(m);
			for (p in poo) if (p != null) gameGraphics.mid.removeChild(p);
			poo = [];
			shurikens.clear();
			cookie = null;

			player.customx = player.x = tama.x;
			player.customy = player.y = tama.y;
			player.cx = tama.cx;
			player.cy = tama.cy;
			target.x = -1;
			target.y = -1;

			var newLevel:LevelData = loadLevel(levelString(0));
			updateWalls(newLevel);

			tempDO = new DisplayObject("thanks");
			tempDO.x = toAbsoluteX(0);
			tempDO.y = toAbsoluteY(19);
			tempDO.__align = DisplayObject.ALIGN_TOPLEFT;
			gameGraphics.mid.unshiftChild(tempDO);
		} else {
			// Clear current level
			Spawner.clear();
			gameGraphics.mid.removeChild(cookie);
			for (m in monsters.copy()) killMonster(m);
			for (p in poo) gameGraphics.mid.removeChild(p);
			poo = [];
			shurikens.clear();

			if (player != null) {
				player.customx = player.x = tama.x;
				player.customy = player.y = tama.y;
				player.cx = tama.cx;
				player.cy = tama.cy;
			}
			target.x = -1;
			target.y = -1;

			if (currLevel == 3) {
				gameGraphics.fade(function() {
					tempDO = new DisplayObject("howto");
					tempDO.x = toAbsoluteX(0);
					tempDO.y = toAbsoluteY(0);
					tempDO.__align = DisplayObject.ALIGN_TOPLEFT;
					gameGraphics.top.addChild(tempDO);

					state = WAIT_INPUT(INSTRUCTIONS_READ);
				});

				// Wait for the fading to end
				state = CINEMATIC(4);
				
				return;
			}

			var newLevel:LevelData = loadLevel(levelString(currLevel));
			updateWalls(newLevel);

			if (currLevel == 2) {
				state = CINEMATIC(1);
				cinematic_start = frame;
			}
		}
	}

	@:access(k.lib.graphics.DisplayObject)
	private function loadLevel(n:String):LevelData {
		var level:LevelData = {
			walls: []
		};
		var bd:BitmapData = Assets.getBitmapData(n);
		var width:Int = bd.width;
		var height:Int = bd.height;

		for (x in 0...width) {
			level.walls[x] = [];

			for (y in 0...height) {
				level.walls[x][y] = false;
				var c:CellColor = cast bd.getPixel(x, y);

				switch (c) {
					case wall: level.walls[x][y] = true;
					case player: player = new Player(toAbsoluteX(x), toAbsoluteY(y)); gameGraphics.midtop.addChild(player);
					case tama: tama = new TamaEntity(toAbsoluteX(x), toAbsoluteY(y)); gameGraphics.midtop.addChild(tama);
					case spawn: Spawner.add(x, y);
					case spawn_unique: 
						var monster:Monster = new Monster(x, y);
						monsters.push(monster);
						gameGraphics.midtop.addChild(monster);

					case cookie:
						cookie = new DisplayObject("cookie");
						cookie.__align = DisplayObject.ALIGN_CENTER;
						cookie.cx = x;
						cookie.cy = y;
						cookie.x = toAbsoluteX(x);
						cookie.y = toAbsoluteY(y);
						gameGraphics.mid.addChild(cookie);

					case empty: // Nothing to do =)
				}
			}
		}

		Spawner.process();
		return level;
	}

	private function updateWalls(level:LevelData, ?loadAll:Bool = false):Void {
		if (loadAll) {
			pathfinder.resetCollisions();
			emptyWalls();
			wallInfo = level.walls.copy();

			for (x in 0...level.walls.length) {
				for (y in 0...level.walls[x].length) {
					var val:Bool = level.walls[x][y];
					pathfinder.setCollision(x, y, val);
					if (!val) continue;

					var wall:DisplayObject = new DisplayObject("wall");
					wall.x = x * CELL_SIZE + PLATEAU_X;
					wall.y = y * CELL_SIZE + PLATEAU_Y;
					wall.cx = x;
					wall.cy = y;
					gameGraphics.mid.addChild(wall);
				}
			}
		} else {
			newWalls = [];
			oldWalls = [];
			for (x in 0...level.walls.length) {
				for (y in 0...level.walls[x].length) {
					var val1:Bool = level.walls[x][y];
					var val2:Bool = wallInfo[x][y];
					pathfinder.setCollision(x, y, val1);

					if (val1 == val2) continue;
					else if (val1) newWalls.push({x: x, y: y});
					else oldWalls.push({x: x, y: y});
				}
			}

			wallInfo = level.walls.copy();
			WALLS_PER_ITERATION = Std.int(Math.max(newWalls.length, oldWalls.length) / 20) + 1;
			newWalls.sort(function(a:Point,b:Point):Int { return Std.int(Math.random() * 50) - 100; });
			oldWalls.sort(function(a:Point,b:Point):Int { return Std.int(Math.random() * 50) - 100; });
		}
	}

	private function emptyWalls():Void {
		if (walls.length == 0) return;

		for (w in walls) if (gameGraphics.mid.contains(w)) gameGraphics.mid.removeChild(w);
	}
}