package k.ld31;

import flash.display.BitmapData;
import flash.display.Sprite;
import flash.filters.DropShadowFilter;
import flash.text.TextField;
import flash.text.TextFormat;
import flash.text.TextFormatAlign;
import flash.text.TextFieldAutoSize;

import k.lib.graphics.Assets;
import k.lib.graphics.DisplayObject;
import k.lib.graphics.RenderEngine;
import k.lib.graphics.Shake;
import k.lib.utils.Color;

import mt.deepnight.Bresenham;

using StringTools;

enum ChatSection {
	FIXED;
	RANDOM;
	MONSTERS;
	HIT;
	PLAYER_HIT;
}

class TamaEntity extends DisplayObject {
	private static inline var FOOD_MAGNET_DISTANCE:Float = 80;
	private static inline var MIN_PLAYER_DISTANCE:Float = 40;
	private static inline var MAX_PLAYER_DISTANCE:Float = 180;
	private static inline var TEXT_DELAY:Int = 200;
	private static inline var SPEED:Float = 1.6;
	public static inline var MAX_LIFE:Int = 5;

	public var life:Int;
	public var lastText:Int;
	public var magnet_on:Bool;
	private static var text_random:Array<Int>;
	private static var text_monster:Array<Int>;
	public static var text_hit:Array<Int>;
	public static var text_playerhit:Array<Int>;

	public static function prepareTexts(render:RenderEngine):Void {
		// Prepare text format
		var s:Sprite;
		var bd:BitmapData;
		var text:TextField = new TextField();
		text.selectable = false;
		text.autoSize = TextFieldAutoSize.NONE;
		text.embedFonts = true;
		text.width = 200;
		text.defaultTextFormat = new TextFormat(Assets.getFontName("FredokaOne"), 11, 0xFFFFFF, false, false, TextFormatAlign.CENTER);
		text.filters = [new DropShadowFilter(0, 0, 0x756115, 1, 5, 5, 2, 1)];

		var texts:Array<String> = Assets.getText("txt/chat.txt").split("\n");
		var section:ChatSection = FIXED;
		text_random = [];
		text_monster = [];
		text_hit = [];
		text_playerhit = [];
		for (i in 0...texts.length) {
			var t:String = texts[i];
			if (t.startsWith("#")) {
				if (t.indexOf("RANDOM") > 0) section = RANDOM;
				else if (t.indexOf("MONSTER") > 0) section = MONSTERS;
				else if (t.indexOf("PHIT") > 0) section = PLAYER_HIT;
				else if (t.indexOf("HIT") > 0) section = HIT;
				continue;
			}

			// Save text
			if (section == RANDOM) text_random.push(i);
			else if (section == RANDOM || section == MONSTERS) text_monster.push(i);
			else if (section == HIT) text_hit.push(i);
			else if (section == PLAYER_HIT) text_playerhit.push(i);

			s = new Sprite();
			text.text = t;
			s.addChild(text);

			bd = new BitmapData(200, 14, true, 0x00FFFFFF);
			bd.draw(s);
			render.add(bd, 'text_$i');
		}
	}

	public function new(x:Int, y:Int) {
		super("tama");
		customx = this.x = x;
		customy = this.y = y;
		has_update = true;
		life = MAX_LIFE;
		radius = 10;
		__align = DisplayObject.ALIGN_CENTER;
		magnet_on = false;

		shakingColorFilter = Shake.FILTER_SOFT;
	}

	@:access(k.lib.graphics.DisplayObject)
	public function talk(?t:Int = -1):Void {
		lastText = Tama.frame;
		if (t == -1) {
			// RANDOM
			if (Spawner.count > 0) t = text_monster[Std.int(Math.random() * text_monster.length)];
			else t = text_random[Std.int(Math.random() * text_random.length)];
		}

		var txt:TamaDialog = new TamaDialog(t, x, y - 18);
	}

	override private inline function onUpdate():Void {
		if (!Tama.locked) {
			var lastcx:Int = cx;
			var lastcy:Int = cy;
			cx = Tama.toCellX(customx);
			cy = Tama.toCellY(customy);

			var targetAcquired:Bool = false;
			var target:DisplayObject = Tama.instance.cookie;
			if (target != null) {
				if (Bresenham.checkFatLine(cx, cy, target.cx, target.cy, Tama.canPass)) {
					var xDistance:Float = target.x - customx;
					var yDistance:Float = target.y - customy;
					var dist:Float = Math.sqrt(yDistance*yDistance + xDistance*xDistance);

					if (dist < FOOD_MAGNET_DISTANCE) {
						if (!magnet_on) talk(4);
						magnet_on = true;

						if (dist <= SPEED) {
							customx = target.x;
							customy = target.y;

							talk(3);
							Tama.instance.nextLevel();
						} else {
							var radian:Float = Math.atan2(yDistance, xDistance);
							customx += Math.cos(radian) * SPEED;
							customy += Math.sin(radian) * SPEED;
						}

						x = Std.int(customx);
						y = Std.int(customy);
						targetAcquired = true;
					}
				}
			}

			// No :
			if (!targetAcquired) {
				magnet_on = false;
				target = Tama.instance.player;
				if (Bresenham.checkThinLine(cx, cy, target.cx, target.cy, Tama.canPass)) {
					var xDistance:Float = target.x - customx;
					var yDistance:Float = target.y - customy;
					var dist:Float = Math.sqrt(yDistance*yDistance + xDistance*xDistance);

					if (dist < MAX_PLAYER_DISTANCE && dist > MIN_PLAYER_DISTANCE) {
						var path = Tama.instance.pathfinder.getPath({x: cx, y: cy}, {x: Tama.toCellX(target.x), y: Tama.toCellY(target.y)});
						// path = Tama.instance.pathfinder.smooth(path);
						if (path.length > 1 && path[0].x == cx && path[0].y == cy) path.shift();
						
						var dest = path[0];
						// Do not go all the way to the target
						if (dest != null && path.length > 1) {
							var xDistance:Float = Tama.toAbsoluteX(dest.x) - customx;
							var yDistance:Float = Tama.toAbsoluteY(dest.y) - customy;

							if (Math.sqrt(yDistance*yDistance + xDistance*xDistance) < SPEED) {
								customx = Tama.toAbsoluteX(dest.x);
								customy = Tama.toAbsoluteY(dest.y);
							} else {
								var radian:Float = Math.atan2(yDistance, xDistance);
								customx += Math.cos(radian) * SPEED;
								customy += Math.sin(radian) * SPEED;
							}
							x = Std.int(customx);
							y = Std.int(customy);
						}
					}
				}
			}

			// No => Idle some way
			if (!targetAcquired) {
				// TODO
			}

			if (Tama.frame - lastText > TEXT_DELAY && Math.random() > 0.95) talk(-1);

			if (Math.random() * 1000 < 2) poo();
			else if (Tama.instance.cookie == null && Math.random() * 100 < 1) poo();
		}
	}

	private function poo():Void {
		var i:Tama = Tama.instance;
		var poo:DisplayObject = new DisplayObject("poo");
		poo.x = Tama.toCellX(x) * Tama.CELL_SIZE + Tama.PLATEAU_X;
		poo.y = Tama.toCellY(y) * Tama.CELL_SIZE + Tama.PLATEAU_Y;
		i.poo.push(poo);
		i.gameGraphics.mid.addChild(poo);
		talk(2);
	}

}
